using System;
using System.Collections.Generic;
using System.Text;
using InfoQ.AspectsOfDMM.Domain;

namespace InfoQ.AspectsOfDMM
{
    class Program
    {
        static void Main(string[] args)
        {
            //Test version 1 of the application
            TestVersion1();

            //Test version 2 of the application
            TestVersion2();

            //Test version 3 of the application
            TestVersion3();

            //Test version 4 of the application
            TestVersion4();

            Console.ReadLine();
        }

        /// <summary>
        /// In the first version all the infrastructure code
        /// has been placed in the domain model. The problem 
        /// with this is that in the long run, our domain model
        /// classes risk becoming overloaded with infrastructure 
        /// code, making them harder to use and maintain. 
        /// </summary>
        private static void TestVersion1()
        {
            Console.WriteLine("Running test 1:");
            Console.WriteLine("-------------------------------------");

            Console.WriteLine("Creating object...");

            //Create a new employee object
            Version1.Employee employee = new Version1.Employee();

            //Write out the dirty and lazy loaded status
            //(The object should not be dirty nor loaded yet)
            WriteObjectStatus(employee);

            Console.WriteLine("Reading from property...");

            //Read from the name property.
            //(This should trigger the lazy loading)
            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded but not dirty)
            WriteObjectStatus(employee);

            Console.WriteLine("Writing to property...");

            //Write to the first name property
            //(This should cause the object to
            //become dirty)
            employee.Name = "Bob";

            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded and dirty)
            WriteObjectStatus(employee);

            Console.WriteLine();
        }


        /// <summary>
        /// In the second version the infrastructure 
        /// code has been moved out from the domain model and 
        /// into a common base class and "proxy" subclasses (one
        /// per domain model class). 
        /// </summary>
        private static void TestVersion2()
        {
            Console.WriteLine("Running test 2:");
            Console.WriteLine("-------------------------------------");

            Console.WriteLine("Creating object...");

            //Create a new employee object
            Employee employee = new Version2.EmployeeProxy();

            //Write out the dirty and lazy loaded status
            //(The object should not be dirty nor loaded yet)
            WriteObjectStatus(employee);

            Console.WriteLine("Reading from property...");

            //Read from the name property.
            //(This should trigger the lazy loading)
            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded but not dirty)
            WriteObjectStatus(employee);

            Console.WriteLine("Writing to property...");

            //Write to the first name property
            //(This should cause the object to
            //become dirty)
            employee.Name = "Bob";

            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded and dirty)
            WriteObjectStatus(employee);

            Console.WriteLine();
        }

        /// <summary>
        /// In the third version the actual logic in the 
        /// infrastructure has been moved out of the proxy
        /// subclasses and into mixin and interceptor classes.
        /// The proxy subclasses, that now essentially contain
        /// only "boilerplate" code, only forwards the calls
        /// to the interceptors and mixins. 
        /// </summary>
        private static void TestVersion3()
        {
            Console.WriteLine("Running test 3:");
            Console.WriteLine("-------------------------------------");

            Console.WriteLine("Creating object...");

            //Create a new employee object
            Employee employee = Version3.Factory.CreateEmployee();

            //Write out the dirty and lazy loaded status
            //(The object should not be dirty nor loaded yet)
            WriteObjectStatus(employee);

            Console.WriteLine("Reading from property...");

            //Read from the name property.
            //(This should trigger the lazy loading)
            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded but not dirty)
            WriteObjectStatus(employee);

            Console.WriteLine("Writing to property...");

            //Write to the first name property
            //(This should cause the object to
            //become dirty)
            employee.Name = "Bob";

            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded and dirty)
            WriteObjectStatus(employee);

            Console.WriteLine();
        }

        /// <summary>
        /// In the fourth version we use an 
        /// Aspect Oriented Programming framework
        /// (NAspect, http://www.puzzleframework.com)
        /// to generate the proxy subclasses for us.
        /// </summary>
        private static void TestVersion4()
        {
            Console.WriteLine("Running test 4:");
            Console.WriteLine("-------------------------------------");

            Console.WriteLine("Creating object...");

            //Create a new employee object
            Version4.Domain.Employee employee = Version4.Factory.CreateEmployee();

            //Write out the dirty and lazy loaded status
            //(The object should not be dirty nor loaded yet)
            WriteObjectStatus(employee);

            Console.WriteLine("Reading from property...");

            //Read from the name property.
            //(This should trigger the lazy loading)
            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded but not dirty)
            WriteObjectStatus(employee);

            Console.WriteLine("Writing to property...");

            //Write to the first name property
            //(This should cause the object to
            //become dirty)
            employee.Name = "Bob";

            Console.WriteLine(
                string.Format(
                "The employee's name is {0}",
                employee.Name));

            //Write out the dirty and lazy loaded status
            //(The object should now be loaded and dirty)
            WriteObjectStatus(employee);

            Console.WriteLine();
        }

        private static void WriteObjectStatus(object obj)
        {
            Console.WriteLine(
                string.Format(
                "Current object dirty status: {0}", 
                ((IDirty) obj).Dirty));
            Console.WriteLine(
                string.Format(
                "Current object loaded status: {0}",
                ((ILazy)obj).Loaded));
        }
    }
}
